module Redundancy (optimizeRedundancy) where
import AST

data DState =
  DS { dsI  :: (Bool, Depth)
     , dsB  :: (Bool, Depth)
     , dsTT :: (Bool, Depth)
     , dsS  :: (Bool, Depth)
     , dsEM :: (Bool, [Depth])
     , dsU  :: (Int, [Depth])
     , dsColor :: (Maybe Color, Depth)
     , dsSize  :: (Maybe Int, Depth)
     }
type Depth = Int

optimizeRedundancy :: [ADoc] -> [ADoc]
optimizeRedundancy = concatMap (fst . remUnused 1 initDState)

remUnused :: Depth -> DState -> ADoc -> ([ADoc],[Depth])
remUnused depth st doc =
  case doc of
    Text _   -> ([doc], whatAffectsText st)
    Space _  -> ([doc], whatAffectsSpace st)
    Tag t ks -> let st' = updateDState st t depth
                    results = map (remUnused (depth+1) st') ks
                    didn'tUseMe (_,deps) = (take 1 deps) /= [depth]
                    (l,ys) = span didn'tUseMe results
                    (r,m)  = span didn'tUseMe (reverse ys)
                    left   = concatMap fst l
                    right  = concatMap fst (reverse r)
                    mid    = concatMap fst (reverse m)
                    retlist = dropWhile (==depth) (mergeAll (map snd results))
                in if null mid then (left, retlist)
                   else ((left ++ Tag t mid : right), retlist)

whatAffectsText :: DState -> [Depth]
whatAffectsText DS { dsI=(_, i), dsB=(_, b), dsTT=(_, tt), dsS=(_, s)
                   , dsEM=(_, em), dsU=(_, u), dsColor=(_, c), dsSize=(_, z)
                   } = revsort ([i,b,tt,s,c,z]++em++u)
                   -- many of these will probably be zeroes (this is OK)

whatAffectsSpace :: DState -> [Depth]
whatAffectsSpace DS { dsTT=(_,tt), dsU=(ulevel, u), dsColor=(_, c), dsSize=(_, z) } =
  if ulevel > 0 then revsort ([tt,c,z]++u)
                else revsort (tt:z:u)

revsort [] = []
revsort xs = let m = maximum xs in m : revsort (filter (/=m) xs)

updateDState :: DState -> Tag -> Depth -> DState
updateDState st tag depth =
  case tag of
    PL -> (setI False . setB False . setTT False . setS False . clearEM . clearU) st
    I                -> (setI True) st
    B                -> (setB True) st
    TT               -> (setTT True) st
    S  | fst (dsS st)-> st
       | otherwise   -> (setS True . clearEM) st
    EM | fst (dsS st)-> st
       | otherwise   -> let (emsetting,ds) = dsEM st
                        in st { dsEM = (not emsetting, depth:ds) }
    U  | fst (dsU st) >= 3 -> st
       | otherwise   -> let (ulevel,ds) = dsU st 
                        in st { dsU = (ulevel+1, depth:ds) }
    Color c          -> (setColor c) st
    Size z           -> (setSize z) st
  where
    setI  x st = if fst (dsI st)  /= x then st { dsI  = (x,depth) } else st
    setB  x st = if fst (dsB st)  /= x then st { dsB  = (x,depth) } else st
    setTT x st = if fst (dsTT st) /= x then st { dsTT = (x,depth) } else st
    setS  x st = if fst (dsS st)  /= x then st { dsS  = (x,depth) } else st
    clearEM st = st { dsEM = (False,[depth]) }
    clearU st  = st { dsU = (0,[depth]) }
    setColor c st = if fst (dsColor st) /= Just c then st { dsColor = (Just c, depth) } else st
    setSize z st = if fst (dsSize st) /= Just z then st { dsSize = (Just z, depth) } else st

initDState :: DState
initDState =
  DS { dsI  = (False, 0)
     , dsB  = (False, 0)
     , dsTT = (False, 0)
     , dsS  = (False, 0)
     , dsEM = (False, [])
     , dsU  = (0, [])
     , dsColor = (Nothing, 0)
     , dsSize  = (Nothing, 0)
     }

mergeAll :: [[Depth]] -> [Depth]
mergeAll = foldr merge []
merge xs [] = xs
merge [] ys = ys
merge (x:xs) (y:ys) =
  case compare x y of
    GT -> x : merge xs (y:ys)
    EQ -> x : merge xs ys
    LT -> y : merge (x:xs) ys
    

