module Rewrite(runRewrite,samePrefix, makePrefix) where

import AST  
import List ((\\))


data State =  MS { hasChanged  :: !Bool
		 , count       :: !Int 
		 }

initialState = MS { hasChanged = False
		  , count = 0 
		  }

maxIterations = 20 :: Int

-- interface to the rest of the world:
runRewrite :: Doc -> Doc
runRewrite d = fst $ unE (rewriteFP d) [] initialState

rewriteFP :: Doc -> Env Doc
rewriteFP d = 
	do d <- rewrite (plShort d)
	   b <- isDone 
	   if b then return d
		else do c <- getCount
		        if c > maxIterations
 			   then return d
			   else  do incCount 
			   	    resetChanged
				    d' <- rewrite d
				    rewriteFP d'

-- Environment + State Monad:
newtype Env a = E ([Tag] -> State  ->  (a,State))
unE (E f) = f

instance Monad Env where
  return x     = E (\env s -> (x,s))
  (E f) >>= k  = 
	E (\env s -> let (r,s') = (f env s)
			 E f' = k r
		     in f' env s' )
		

isUnderTag :: Tag -> Env Bool
isUnderTag t = E (\env s -> (t `elem` env,s))

inExtendedEnv :: Tag -> Env a -> Env a
inExtendedEnv tag m = E (\env s  -> unE m (tag : env) s)

plEnv :: Env a -> Env a
plEnv m = E (\env s -> unE m (PL : (filter (`notElem` [I, B, EM, TT, U, S]) env)) s)

changed :: Env ()
changed = E (\env s -> ((), s {hasChanged = True })) 

resetChanged :: Env ()
resetChanged = E (\env s -> ((), s {hasChanged = False}))

isDone :: Env Bool
isDone = E (\env s -> (not $ hasChanged s, s) )

isWhiteSpaceMergeable :: Env Bool
isWhiteSpaceMergeable = E (\env s -> 
	(not $ any pred env, s) ) 
		where	pred TT = True
			pred U  = True
			pred (Size _ )= True
			pred PL       = True
			pred _        = False
			
incCount :: Env ()
incCount = E (\env s -> ((), s {count = count s + 1 }))

getCount :: Env Int
getCount = E (\env s -> (count s, s))

-- rewrite engine:
rewrite :: Doc -> Env Doc
rewrite (Text s  : r) = do r' <- rewrite r
                           return (Text s : r')
rewrite (Space s : r) = do r' <- rewrite r
			   return (Space s : r')
rewrite ((Tag PL ds1):(Tag PL ds2):rest) = 
	do ds1' <- plEnv $ (rewrite ds1)
	   ds2' <- plEnv $ (rewrite ds2)
	   changed
	   rewrite $ (Tag PL (plShort (ds1' ++ ds2'))) : rest
rewrite (x@(Tag t1 ds1):Space s: y@(Tag t2 ds2):rest) = 
	do 
	 	ds1'' <- (if t1 == PL then plEnv else inExtendedEnv t1) $ rewrite ds1
		ds2'' <- (if t2 == PL then plEnv else inExtendedEnv t2) $ rewrite ds2
		let ds1' = if t1 == PL then plShort ds1'' else ds1''
		    ds2' = if t1 == PL then plShort ds2'' else ds2''
		b <- isWhiteSpaceMergeable
		if b
		 then let (common,lefttags,righttags,leftchild,rightchild) = 
				samePrefixNoTT (Tag t1 ds1') (Tag t2 ds2') ([],[]) 
		      in case common of 
  				[] -> do rest' <- rewrite (Space s :(Tag t2 ds2'): rest)
					 return $ (Tag  t1 ds1'):rest'
				_  -> do changed
					 newT <- return $ makePrefix common $ (makePrefix lefttags leftchild) ++ [Space s] ++ (makePrefix righttags rightchild)
					 rewrite $ newT ++ rest
					 -- or rewrite rest >>= \rest' -> return $ newT ++ rest' 
		 else  do rest' <- rewrite (Space s :(Tag t2 ds2'): rest)
			  return $ (Tag  t1 ds1'):rest'
rewrite (x@(Tag t1 ds1):y@(Tag t2 ds2):rest) =  
 case candidate x y of 
    Just zs -> changed >> rewrite (zs ++ rest)
    _ -> 
	do
   	   ds1'' <- (if t1 == PL then plEnv else inExtendedEnv t1) $ rewrite ds1
  	   ds2'' <- (if t2 == PL then plEnv else inExtendedEnv t2) $ rewrite ds2
	   let ds1' = if t1 == PL then plShort ds1'' else ds1''
	       ds2' = if t1 == PL then plShort ds2'' else ds2''
	   let w@(common,lefttags,righttags,leftchild,rightchild) = samePrefix (Tag t1 ds1') (Tag t2 ds2') ([], [])
	   case common of 
		[] -> do rest' <- rewrite ((Tag t2 ds2'):rest)
		         return $ (Tag t1 ds1'):rest'
	 	_ -> 	do changed
		 	   let newT = makePrefix common $ (makePrefix lefttags leftchild  ) ++
			 		                  (makePrefix righttags rightchild)
			   rewrite $ newT ++ rest 
rewrite (Tag t ds : r)
  | t == PL        = do ds' <- plEnv $ rewrite ds
			r'  <- rewrite r
		 	return $ (Tag PL (plShort ds')) : r'
  | True           = -- no optimization applies, continue:
  		     do ds' <- inExtendedEnv t $ rewrite ds
                        r'  <- rewrite r
                        return (Tag t ds' : r')
rewrite []            = return []


samePrefix :: ADoc -> ADoc -> ([Tag],[Tag]) -> ([Tag],[Tag],[Tag], [ADoc], [ADoc])
samePrefix (t1 @ (Tag PL _ )) t2 (l1,l2) = (intersection, l1 \\ intersection, l2 \\ intersection, [t1],[t2])
	where intersection = l1 `intersect` l2
samePrefix t1 (t2 @ (Tag PL _ )) (l1,l2) = (intersection, l1 \\ intersection, l2 \\ intersection, [t1],[t2])
	where intersection = l1 `intersect` l2
samePrefix (Tag t1 [tr1]) (Tag t2 [tr2]) (l1,l2) = samePrefix tr1 tr2 (t1:l1,t2:l2)
samePrefix (Tag t1 ts1)   (Tag t2 ts2)   (l1,l2) = (intersection, (t1 : l1) \\ intersection, 
                                                                  (t2 : l2) \\ intersection,
			                            ts1, ts2)
	  where intersection = (t1:l1) `intersect` (t2:l2)
samePrefix (Tag t1 [ts1]) t2              (l1,l2) = samePrefix ts1 t2 (t1:l1,l2)
samePrefix t1             (Tag t2 [ts2])  (l1,l2) = samePrefix t1 ts2 (l1,t2:l2)
samePrefix t1             t2              (l1,l2) = (intersection, l1 \\ intersection, l2 \\ intersection,[t1],[t2])
	                                          where intersection = l1 `intersect` l2


samePrefixNoTT :: ADoc -> ADoc -> ([Tag],[Tag]) -> ([Tag],[Tag],[Tag], [ADoc], [ADoc])
samePrefixNoTT (t1 @ (Tag PL _ )) t2 (l1,l2) = (intersection, l1 \\ intersection, l2 \\ intersection, [t1],[t2])
	where intersection = l1 `intersect` l2
samePrefixNoTT t1 (t2 @ (Tag PL _ )) (l1,l2) = (intersection, l1 \\ intersection, l2 \\ intersection, [t1],[t2])
	where intersection = l1 `intersect` l2
samePrefixNoTT (t1 @ (Tag TT _ )) t2 (l1,l2) = (intersection, l1 \\ intersection, l2 \\ intersection, [t1],[t2])
	where intersection = l1 `intersect` l2
samePrefixNoTT t1 (t2 @ (Tag TT _ )) (l1,l2) = (intersection, l1 \\ intersection, l2 \\ intersection, [t1],[t2])
	where intersection = l1 `intersect` l2
samePrefixNoTT (Tag t1 [tr1]) (Tag t2 [tr2]) (l1,l2) = samePrefixNoTT tr1 tr2 (t1:l1,t2:l2)
samePrefixNoTT (Tag t1 ts1)   (Tag t2 ts2)   (l1,l2) = (intersection, (t1 : l1) \\ intersection, 
                                                                  (t2 : l2) \\ intersection,
			                            ts1, ts2)
	  where intersection = (t1:l1) `intersect` (t2:l2)
samePrefixNoTT (Tag t1 [ts1]) t2              (l1,l2) = samePrefixNoTT ts1 t2 (t1:l1,l2)
samePrefixNoTT t1             (Tag t2 [ts2])  (l1,l2) = samePrefixNoTT t1 ts2 (l1,t2:l2)
samePrefixNoTT t1             t2              (l1,l2) = (intersection, l1 \\ intersection, l2 \\ intersection,[t1],[t2])
	                                          where intersection = l1 `intersect` l2


-- bag intersection:
intersect []     ys = []
intersect (x:xs) ys
   | x `elem` ys    = x : intersect xs (ys \\ [x])
   | True           = intersect xs ys

		
makePrefix []     ds = ds
makePrefix [tag]  ds = [Tag tag ds]
makePrefix (t:ts) ds = [Tag t (makePrefix ts ds)]


candidate t1 (Tag t [t2,t3]) = 
  let (common,lefttag,righttag,leftchild,rightchild) = samePrefix t1 t2 ([],[])
      len1 = length  $  (common++lefttag++[t]++common++righttag)
      len2 = length  $  (common ++ lefttag ++ [t,t])

  in
   if len1 > len2 
    then 
	case common of 
		[] -> Nothing
		r@(x:y:rest) -> 
			Just $ (makePrefix common $ (makePrefix lefttag leftchild) ++ 
				[(Tag t (makePrefix righttag rightchild)) ]) ++
				 [Tag t [t3] ]
		_ -> Nothing
    else Nothing
candidate _ _ = Nothing

plShort (t1 : Text s : t2 : rest) = 
   let (common', lefttag',righttag',leftchild,rightchild) = samePrefix t1 t2 ([],[])
       common = filter (not . isSizeOrColor) common' 
       lefttag  = lefttag'  ++ (filter isSizeOrColor common')
       righttag = righttag' ++ (filter isSizeOrColor common')
   in     
	case common of 
		_:_:_ -> (makePrefix common $ (makePrefix lefttag leftchild) ++ [ Tag PL [Text s]]  ++ (makePrefix righttag rightchild) ) ++ (plShort rest)
		_     -> t1 : Text s : t2 : (plShort $  rest)
		where isSizeOrColor (Size _) = True
		      isSizeOrColor (Color _) = True
		      isSizeOrColor _ = False
plShort [] = []
plShort (x:xs) = x : plShort xs
